/*==============================================================================*/
/*  Copyright (C) 2022, Heartland Data inc. All Rights Reserved.                */
/*                                                                              */
/*  Title  :   GPIO 2bit Driver                                                 */
/*  EventID:   Insert Kernelinfo Test Point                                     */
/*  AppVer :   DT+ Ver1.00~                                                     */
/*  FileID :   34b1                                                             */
/*  Version:   2.0                                                              */
/*  Author :   HLDC                                                             */
/*==============================================================================*/

/*==============================================================================*/
/*  Please customize the code for your environment.                             */
/*==============================================================================*/

/*==============================================================================*/
/*  Desc:   Header for PORT Control                                             */
/*==============================================================================*/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <linux/sched.h>
#include <asm/io.h>
#include <linux/version.h>
#include <linux/delay.h>

#define LICENSE "GPL v2"

/*==============================================================================*/
/*  Macro:  DT_UINT                                                             */
/*  Desc:   Please change Test Point argument type for DT+ Project setting.     */
/*==============================================================================*/
#define	DT_UINT unsigned int

/*==============================================================================*/
/*  Macro:  PROC_NAME                                                           */
/*  Desc:   Register Proc File name for Linux.                                  */
/*==============================================================================*/
#define	PROC_NAME "dtproc"

/*==============================================================================*/
/*  Macro:  DT_USE_PREDICT_TP                                                   */
/*  Desc:   Please set 1 if use Predict Test Point (DT+ Ver2.00 or later Only)  */
/*==============================================================================*/
#define DT_USE_PREDICT_TP 0

/*==============================================================================*/
/*  Struct:  DT_DATA                                                            */
/*  Desc:    Test Point Data Struct                                             */
/*==============================================================================*/
struct DT_DATA {
	char tp_func;
	DT_UINT addr;
	DT_UINT dat;
	DT_UINT size;
	unsigned char value[256];
};

/*==============================================================================*/
/*  Desc:    DT_DATA.tp_func ID                                                 */
/*==============================================================================*/
#define TP_BusOut		'2'
#define TP_MemoryOutput		'3'
#define TP_BusOutFast		'4'
#define TP_MemoryOutputFast		'5'
#define TP_EventTrigger		'6'
#define TP_EventTriggerFast		'7'
#define TP_EventTrigger32		'8'
#define TP_EventTrigger32Fast		'9'
#if DT_USE_PREDICT_TP
#define TP_PredictBusOut		'A'
#endif

/*==============================================================================*/
/*  Desc:   Test Point Parametaer                                               */
/*==============================================================================*/
#define		DT_VARIABLE_BIT			0x02
#define		DT_EVTTRG_BIT			0x08
#define		DT_VARIABLE_FAST_BIT		0x01
#define		DT_EVTTRG_FAST_BIT		0x02

/*==============================================================================*/
/*  Macro:  DT_ADD_CPU_INFO                                                     */
/*  Desc:   Please change Test Point argument type for DT+ Project setting.     */
/*==============================================================================*/
#define DT_ADD_CPU_INFO		0

/*==============================================================================*/
/*  Macro:  DT_USE_SPINLOCK                                                     */
/*  Desc:   Please set 1, when use spin_lock_irqsave function.                  */
/*==============================================================================*/
#include <linux/spinlock.h>
static DEFINE_SPINLOCK(tplock);
unsigned long flags = 0;
#define DT_LOCK	spin_lock_irqsave(&tplock, flags)
#define DT_UNLOCK	spin_unlock_irqrestore(&tplock, flags)

/*==============================================================================*/
/*  Desc:   Global Variable of Predict Test Point                               */
/*==============================================================================*/
#if DT_USE_PREDICT_TP
DT_UINT predict_addr = 0;
DT_UINT predict_dat = 0;
#endif

/*==============================================================================*/
/*  Desc:   GPIO Control setting                                                */
/*==============================================================================*/
#define BCM2708_PERI_BASE        0x20000000
#define GPIO_BASE                (BCM2708_PERI_BASE + 0x200000)

#define	CS_BIT		7	/* CS */
#define	CLK_BIT		25	/* CLK */
#define	DAT0_BIT	8	/* DAT0 */
#define	DAT1_BIT	9	/* DAT1 */

void *iop = NULL;

#define	INP_GPIO(g) iowrite32(\
	ioread32( (void*)((int *)iop+(g)/10) ) & ~(7<<(((g)%10)*3)), \
		  (void*)((int *)iop+(g)/10) )
#define OUT_GPIO(g) iowrite32(\
	ioread32( (void*)((int *)iop+(g)/10) ) | (1<<(((g)%10)*3)), \
		  (void*)((int *)iop+(g)/10) )
#define	GPIO_SET(g) iowrite32( 1<<(g), (void*)((int *)iop+7) )
#define	GPIO_CLR(g) iowrite32( 1<<(g), (void*)((int *)iop+10) )
#define	GPIO_GET ioread32((void*)((int *)iop+13))

#define	CLK_CS_HI	{			\
	GPIO_SET(CS_BIT);			\
	GPIO_SET(CLK_BIT);			\
}

#define	CS_HI	{				\
	GPIO_SET(CS_BIT);			\
}
#define	CS_LOW	{				\
	GPIO_CLR(CS_BIT);			\
}

#define	CLK_HI	{				\
	GPIO_SET(CLK_BIT);			\
}

#define	CLK_LOW	{				\
	GPIO_CLR(CLK_BIT);			\
}

#define DATA_OUT(x) {			\
	( (x) & 0x1) ? GPIO_SET(DAT0_BIT) : GPIO_CLR(DAT0_BIT);	\
	( (x) & 0x2) ? GPIO_SET(DAT1_BIT) : GPIO_CLR(DAT1_BIT);	\
	udelay(1);	\
}

static int init = 0;	/* init flag */
static unsigned int start_trace = 0;
/*==============================================================================*/
/*  Func:   _TP_BusPortInit                                                     */
/*  Desc:   Please describe the code to initializes ports.                      */
/*==============================================================================*/
static void _TP_BusPortInit(void)
{
	if (init == 0) {
		iop = ioremap(GPIO_BASE,16);
		if (iop == NULL) {
			printk("dtdrvspi: init err\n");
			return;
		}
		OUT_GPIO(CS_BIT);
		OUT_GPIO(CLK_BIT);
		OUT_GPIO(DAT0_BIT);
		OUT_GPIO(DAT1_BIT);

		CLK_CS_HI;

		init = 1;
		return;
	}
}

/*==============================================================================*/
/*  Func:   _TP_BusOutDrv                                                       */
/*  Desc:   Test Point Output Function.                                         */
/*==============================================================================*/
static inline void _TP_BusOutDrv( DT_UINT addr, DT_UINT data )
{
#if DT_ADD_CPU_INFO
	int cpu_id = 0;
	cpu_id = smp_processor_id();
	if ((cpu_id < 0) || (cpu_id > 3)) {
		cpu_id = 0;
	}
	addr &= ~(0x3 << 18); 
	addr |= cpu_id << 18;
#endif
	CS_LOW;

	DATA_OUT(data >> 14);		/* bit15-14 */
	CLK_LOW;
	DATA_OUT(data >> 12);		/* bit13-12 */
	CLK_HI;
	DATA_OUT(data >> 10);		/* bit11-10 */
	CLK_LOW;
	DATA_OUT(data >> 8);		/* bit9-8 */
	CLK_HI;
	DATA_OUT(data >> 6);		/* bit7-6 */
	CLK_LOW;
	DATA_OUT(data >> 4);		/* bit5-4 */
	CLK_HI;
	DATA_OUT(data >> 2);		/* bit3-2 */
	CLK_LOW;
	DATA_OUT(data);			/* bit1-0 */
	CLK_HI;

	if( addr >= 0x00010000 ){
		DATA_OUT( addr >> 22 );
		CLK_LOW;
		DATA_OUT( addr >> 20 );
		CLK_HI;
		DATA_OUT( addr >> 18 );
		CLK_LOW;
		DATA_OUT( addr >> 16 );
		CLK_HI;
	}
	if( addr >= 0x00000100 ){
		DATA_OUT( addr >> 14 );
		CLK_LOW;
		DATA_OUT( addr >> 12 );
		CLK_HI;
		DATA_OUT( addr >> 10 );
		CLK_LOW;
		DATA_OUT( addr >> 8 );
		CLK_HI;
	}
	DATA_OUT( addr >> 6 );
	CLK_LOW;
	DATA_OUT( addr >> 4 );
	CLK_HI;
	DATA_OUT( addr >> 2 );
	CLK_LOW;
	DATA_OUT( addr );
	CLK_HI;

	CS_HI;
}

/*==============================================================================*/
/*  Func:   _TP_MemoryOutDrv                                                    */
/*  Desc:   Value Output Function.                                              */
/*==============================================================================*/
static inline void _TP_MemoryOutDrv(unsigned char data)
{
	DATA_OUT(data >> 6 );			/* bit7-6 */
	CLK_LOW;
	DATA_OUT(data >> 4 );			/* bit5-4 */
	CLK_HI;
	DATA_OUT(data >> 2 );			/* bit3-2 */
	CLK_LOW;
	DATA_OUT(data);				/* bit1-0 */
	CLK_HI;
}

/*==============================================================================*/
/*  Func:   _TP_BusOutFastDrv                                                   */
/*  Desc:   Test Point Output Function.                                         */
/*==============================================================================*/
static void _TP_BusOutFastDrv(DT_UINT data, DT_UINT bit_num)
{
	CS_LOW;
	switch (bit_num) {
	case 12:
	case 11:
		DATA_OUT((data >> 10) & 0xFF);
		CLK_LOW;
	case 10:
	case 9:
		DATA_OUT((data >> 8) & 0xFF);
		CLK_LOW;
	case 8:
	case 7:
		DATA_OUT((data >> 6) & 0xFF);
		CLK_LOW;
	case 6:
	case 5:
		DATA_OUT((data >> 4) & 0xFF);
		CLK_LOW;
	default:
		DATA_OUT((data) & 0xFF);
		CLK_LOW;
		break;

	}
	CS_HI;
	CLK_HI;
}

#if DT_USE_PREDICT_TP
/*==============================================================================*/
/*  Func:   _TP_PredictBusOutDrv                                                   */
/*  Desc:   Predict Test Point Output Function.                                 */
/*==============================================================================*/
static void _TP_PredictBusOutDrv(void)
{
	CS_LOW;
	DATA_OUT(0x1);
	CLK_LOW;
	CS_HI;
	CLK_HI;
}
#endif

/**********************************kernel TestPoint API**********************************/

/*==============================================================================*/
/*  Func:   _TP_BusOut                                                          */
/*  Desc:   Called by Test Point                                                */
/*==============================================================================*/
#if DT_USE_PREDICT_TP
void _TP_BusOut( DT_UINT addr, DT_UINT dat, DT_UINT p_addr, DT_UINT p_dat )
{
	if (start_trace != 1) {
		return;
	}
	DT_LOCK;
	if (predict_dat != dat || predict_addr != addr) {
		_TP_BusOutDrv(addr, dat);
	} else {
		_TP_PredictBusOutDrv();
	}
	predict_addr = p_addr;
	predict_dat = p_dat;
	DT_UNLOCK;
}
#else
void _TP_BusOut( DT_UINT addr, DT_UINT dat )
{
	if (start_trace != 1) {
		return;
	}
	DT_LOCK;
	_TP_BusOutDrv(addr, dat);
	DT_UNLOCK;
}
#endif
EXPORT_SYMBOL(_TP_BusOut);

/*==============================================================================*/
/*  Func:   _TP_MemoryOutput                                                    */
/*  Desc:   Called by Variable TestPoint                                        */
/*==============================================================================*/
void _TP_MemoryOutput( DT_UINT addr, DT_UINT dat, void *value, DT_UINT size )
{
	int	i;
	unsigned char *p = (unsigned char *)value;

	if (start_trace != 1) {
		return;
	}

	if (size > 256) size = 256;
	DT_LOCK;
	_TP_BusOutDrv((addr | DT_VARIABLE_BIT), dat);
	CS_LOW;
	_TP_MemoryOutDrv(size);
	for (i = 0; i < size; i++) {
		_TP_MemoryOutDrv(p[i]);
	}
	CS_HI;
	DT_UNLOCK;
}
EXPORT_SYMBOL(_TP_MemoryOutput);

/*==============================================================================*/
/*  Func:   _TP_BusOutFast                                                      */
/*  Desc:   Called by Fast Test Point                                           */
/*==============================================================================*/
void _TP_BusOutFast( DT_UINT step, DT_UINT bit_num )
{
	if (start_trace != 1) {
		return;
	}

	step = step << 2;
	bit_num = bit_num + 2;
	DT_LOCK;
	_TP_BusOutFastDrv(step, bit_num);
#if DT_USE_PREDICT_TP
	predict_addr = 0;
	predict_dat = 0;
#endif
	DT_UNLOCK;
}
EXPORT_SYMBOL(_TP_BusOutFast);

/*==============================================================================*/
/*  Func:   _TP_MemoryOutputFast                                                */
/*  Desc:   Called by Variable Fast Test Point                                  */
/*==============================================================================*/
void _TP_MemoryOutputFast( DT_UINT step, DT_UINT bit_num, void *value, DT_UINT size)
{
	int	i;
	unsigned char *p = (unsigned char *)value;

	if (start_trace != 1) {
		return;
	}

	step = (step << 2) + DT_VARIABLE_FAST_BIT;
	bit_num = bit_num + 2;
	if (size > 256) size = 256;
	DT_LOCK;
	_TP_BusOutFastDrv(step, bit_num);
	CS_LOW;
	_TP_MemoryOutDrv(size);
	for (i = 0; i < size; i++) {
		_TP_MemoryOutDrv(p[i]);
	}
	CS_HI;
	DT_UNLOCK;
}
EXPORT_SYMBOL(_TP_MemoryOutputFast);

/*==============================================================================*/
/*  Func:   _TP_BusKernelInfo                                                   */
/*  Desc:   Called by KernelInfo Test Point.                                    */
/*==============================================================================*/
void _TP_BusKernelInfo(DT_UINT addr, DT_UINT dat, struct task_struct *next)
{
	int i;
	unsigned int pid = next->pid;
	unsigned int tgid = next->mm?next->tgid:0;
	unsigned char *p = next->comm;

	if (start_trace != 1) {
		return;
	}

	DT_LOCK;
	_TP_BusOutDrv((addr | DT_VARIABLE_BIT | DT_EVTTRG_BIT), dat);
	CS_LOW;
	_TP_MemoryOutDrv(pid);
	_TP_MemoryOutDrv((pid >> 8));
	_TP_MemoryOutDrv((pid >> 16));
	_TP_MemoryOutDrv((pid >> 24));
	_TP_MemoryOutDrv(20);
	_TP_MemoryOutDrv(tgid);
	_TP_MemoryOutDrv((tgid >> 8));
	_TP_MemoryOutDrv((tgid >> 16));
	_TP_MemoryOutDrv((tgid >> 24));
	for (i = 0; i < 16; i++) {
		_TP_MemoryOutDrv(p[i]);
	}
	CS_HI;
	DT_UNLOCK;
}
EXPORT_SYMBOL(_TP_BusKernelInfo);

/********************************************************************************/

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
static ssize_t dt_proc_write(struct file *filp, const char __user *buf, size_t len, loff_t *data)
#else
static int dt_proc_write(struct file *filp, const char *buf, unsigned long len, void *data)
#endif
{
	struct DT_DATA dt_data;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 4, 0))
	char tmp;

	if (copy_from_user(&tmp, buf, 1)) {
		return len;
	}
	switch (tmp) {
#else
	switch (buf[0]) {
#endif
	case '0':
		start_trace = 0;
		printk( KERN_INFO "[DT] Trace Stop.\n" );
		break;
	case '1':
		start_trace = 1;
		printk( KERN_INFO "[DT] Trace Start.\n" );
		break;
	case TP_BusOut:
		if (copy_from_user(&dt_data, buf, len) == 0) {
			if (start_trace == 1) {
				DT_LOCK;
				_TP_BusOutDrv(dt_data.addr, dt_data.dat);
				DT_UNLOCK;
			}
		}
		break;
	case TP_MemoryOutput:
		if (copy_from_user(&dt_data, buf, len) == 0) {
			_TP_MemoryOutput(dt_data.addr, dt_data.dat, dt_data.value, dt_data.size);
		}
		break;
	case TP_BusOutFast:
		if (copy_from_user(&dt_data, buf, len) == 0) {
			if (start_trace == 1) {
				DT_LOCK;
				_TP_BusOutFastDrv((dt_data.addr << 2), (dt_data.dat + 2));
				DT_UNLOCK;
			}
		}
		break;
	case TP_MemoryOutputFast:
		if (copy_from_user(&dt_data, buf, len) == 0) {
			_TP_MemoryOutputFast(dt_data.addr, dt_data.dat, dt_data.value, dt_data.size);
		}
		break;
#if DT_USE_PREDICT_TP
	case TP_PredictBusOut:
		if (start_trace == 1) {
			DT_LOCK;
			_TP_PredictBusOutDrv();
			DT_UNLOCK;
		}
		break;
#endif
	default:
		printk( KERN_INFO "[DT] Trace Command is 0-5.\n" );
		break;
	}

	return len;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
static ssize_t dt_proc_read(struct file *filp, char __user *buf, size_t count, loff_t *data)
#else
static int dt_proc_read(char *buf, char **start, off_t off, int count, int *eof, void *data)
#endif
{
#if !(LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
	*eof = 1;
#endif
	return count;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
static const struct file_operations dt_proc_fops = {
	.write = dt_proc_write,
	.read = dt_proc_read,
};
#endif

static int __init dt_proc_init(void)
{
	struct proc_dir_entry* entry;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
	entry = proc_create(PROC_NAME, S_IRUGO | S_IWUGO, NULL, &dt_proc_fops);
	if ( !entry ) {
		printk( KERN_ERR "proc entry fail\n" );
		return -EBUSY;
	}
#else
	entry = create_proc_entry(PROC_NAME, S_IRUGO | S_IWUGO, NULL);
	entry->read_proc = dt_proc_read;
	entry->write_proc = dt_proc_write;
#endif
	_TP_BusPortInit();
	printk(KERN_INFO "dtproc load OK\n");
	if(init == 1){
		start_trace = 1;
	}
	return 0;
}

static void __exit dt_proc_exit(void)
{
	remove_proc_entry(PROC_NAME, NULL);
	if (init != 0) {
		start_trace = 0;
		init = 0;
		iounmap(iop);
	}
	printk(KERN_INFO "dtproc unload\n");
}

module_init(dt_proc_init);
module_exit(dt_proc_exit);

MODULE_DESCRIPTION("dtproc");
MODULE_LICENSE(LICENSE);

/********************************************************************************/
